# idownload v1.1 (https://github.com/PCigales/WinSocketTB)
# Copyright © 2025 PCigales
# This program is licensed under the GNU GPLv3 copyleft license (see https://www.gnu.org/licenses)

import sys
import subprocess
import json
import os, os.path
import threading
import time
import msvcrt


try:
  if not (length := int.from_bytes(sys.stdin.buffer.read(4), sys.byteorder)):
    raise
  message = json.loads(sys.stdin.buffer.read(length))
  if 'explorer' in message:
    started = True
    path = message['explorer']
    if os.path.isfile(path):
      subprocess.run('explorer /select,' + path)
    else:
      path = os.path.dirname(path)
      if not os.path.isdir(path):
        raise
      os.startfile(path, 'explore')
    exit(0)
  try:
    from SocketTB import HTTPIDownload, WebSocketDataStore, IDAltSocketGenerator, WebSocketIDClient
  except:
    import importlib.util
    spec = importlib.util.spec_from_file_location('SocketTB', os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(globals().get('__file__', ' '))), os.pardir, 'SocketTB.py')))
    spec.loader.exec_module(sys.modules.setdefault('SocketTB', importlib.util.module_from_spec(spec)))
    from SocketTB import HTTPIDownload, WebSocketDataStore, IDAltSocketGenerator, WebSocketIDClient
  url = message['url']
  file = message['file']
  dfile = file + '.idownload'
  progress = message.get('progress')
  download = HTTPIDownload(url, dfile, headers={k: v for k, v in map(dict.values, message['headers']) if k.lower() != 'range'}, max_workers=message['maxsecs'], section_min=(message['secmin'] * 1048576), file_preallocate=True, file_sparse=message['sparse'], resume=progress)
  if not download:
    raise
  download.start()
  download.wait_progression()
  started = (st := download.progress['status']) != 'aborted'
except:
  started = False
  exit(1)
finally:
  response = json.dumps(started, separators=(',', ':')).encode()
  sys.stdout.buffer.write(len(response).to_bytes(4, sys.byteorder))
  sys.stdout.buffer.write(response)
  sys.stdout.buffer.flush()
  sys.stderr.buffer.write(b' ')
  sys.stderr.buffer.flush()
sys.stdin = open('con', 'r')
sys.stdout = open('con', 'w')
sys.stderr = open('con', 'w')
download.sdid = message['sdid']
download.suspended = not started and progress is not None


class DownloadReportDS(WebSocketDataStore):

  def __init__(self, download):
    super().__init__()
    self.incoming_text_only = True
    self.download = download
    self.client = None

  @getattr(property(), 'setter')
  def progress(self, value):
    value.pop('workers', None)
    self.set_outgoing(0, json.dumps({'sdid': self.download.sdid, 'progress': value}, separators=(',', ':')))

  def add_incoming(self, value):
    if value == 'discard %s' % self.download.sdid:
      self.download.stop()
    elif value == 'suspend %s' % self.download.sdid:
      self.download.suspended = True
      self.download.stop()


def connect(port, download_ds):
  IDSockGen = IDAltSocketGenerator()
  to = 0.2
  while True:
    if (DownloadWSClient := WebSocketIDClient('ws://localhost:%d/report' % port, download_ds, headers={'X-Request-Id': download_ds.download.sdid}, connection_timeout=to, idsocket_generator=IDSockGen)) is None:
      to = 1
      process = subprocess.Popen(('py', os.path.join(os.path.dirname(os.path.abspath(globals().get('__file__', ' '))), 'websocket.py'), str(port)), creationflags=(subprocess.CREATE_BREAKAWAY_FROM_JOB | subprocess.CREATE_NO_WINDOW), stderr=subprocess.PIPE)
      if process.stderr.read(1) != b'0':
        if download_ds.before_shutdown:
          return
        time.sleep(1)
        if download_ds.before_shutdown:
          return
    else:
      download_ds.client = DownloadWSClient
      return


DownloadDS = DownloadReportDS(download)
th = threading.Thread(target=connect, args=(message['port'], DownloadDS))
th.start()

print('Downloading:', url)
print('Size:', download.progress['size'] or '?')
print('into:', dfile)
print()
DownloadDS.progress = download.progress
if started and st != 'completed':
  print('status:', st)
  print('progression: %s' % download.wait_progress_bar(100, 0), end='\b'*118, flush=True)
  while True:
    print('progression: %s' % download.wait_progress_bar(100), end='\b'*118, flush=True)
    progress = download.progress
    if (st := progress['status']) in ('completed', 'aborted'):
      break
    DownloadDS.progress = progress
  print('progression: %s' % download.wait_progress_bar(100))
suspended = download.suspended
print('status: %s%s' % (st, ((' (%s)' % e) if st == 'aborted' and isinstance((e := download.progress['error']), str) else '')))
DownloadDS.progress = download.progress if st != 'aborted' or suspended else {'status': 'aborted', 'size': download.progress['size'], 'downloaded': 0, 'percent': 0, 'error': download.progress['error']}
if st == 'completed':
  while True:
    try:
      os.rename(dfile, file)
      print('\r\nrenamed:', file)
      break
    except:
      if not os.path.isfile(dfile):
        print('\r\nmissing:', dfile)
        break
      try:
        os.remove(file)
      except:
        time.sleep(0.5)
elif not suspended:
  while True:
    try:
      os.remove(dfile)
      break
    except:
      if not os.path.isfile(dfile):
        break
      time.sleep(0.5)

if not message.get('selfclose'):
  while msvcrt.kbhit():
    if msvcrt.getch() == b'\xe0':
      msvcrt.getch()
  print('\r\nPress any key to exit')
  msvcrt.getch()

DownloadDS.before_shutdown = 'end'
th.join()
if DownloadDS.client:
  DownloadDS.client.close(once_data_sent=True, block_on_close=True)


# idownload v1.1 (https://github.com/PCigales/WinSocketTB)
# Copyright © 2025 PCigales
# This program is licensed under the GNU GPLv3 copyleft license (see https://www.gnu.org/licenses)

import sys
import json
import os.path
import threading
import time
import ctypes, ctypes.wintypes


try:
  from SocketTB import WebSocketDataStore, WebSocketRequestHandler, WebSocketIDAltServer
except:
  try:
    import importlib.util
    spec = importlib.util.spec_from_file_location('SocketTB', os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(globals().get('__file__', ' '))), os.pardir, 'SocketTB.py')))
    spec.loader.exec_module(sys.modules.setdefault('SocketTB', importlib.util.module_from_spec(spec)))
    from SocketTB import WebSocketDataStore, WebSocketRequestHandler, WebSocketIDAltServer
  except:
    exit(1)


class DownloadsMonitorDS(WebSocketDataStore):

  def __init__(self, report_datastore):
    super().__init__()
    self.incoming_text_only = True
    self.downloads = {}
    self.report_datastore = report_datastore

  @property
  def progress(self):
    return lambda sdid: None if (ind := self.downloads.get(sdid)) is None else self.outgoing[ind][1]

  @progress.setter
  def progress(self, value):
    progression = json.loads(value)
    if progression['progress']['status'] in ('completed', 'aborted'):
      self.report_datastore.downloaders.pop(progression['sdid'], None)
    self.set_outgoing(self.downloads.setdefault(progression['sdid'], len(self.downloads)), value)

  def add_incoming(self, value):
    self.report_datastore.command = value


class DownloadsReportDS(WebSocketDataStore):

  def __init__(self, server):
    super().__init__()
    self.incoming_text_only = True
    self.server = server
    self.downloaders = {}
    self.monitor_datastore = DownloadsMonitorDS(self)

  @getattr(property(), 'setter')
  def command(self, value):
    self.server.broadcast('report', value)

  def add_incoming(self, value):
    self.monitor_datastore.progress = value


class DownloadsWSRequestHandler(WebSocketRequestHandler):

  def connected_callback(self):
    super().connected_callback()
    try:
      if self.request.getsockname()[0] != self.request.getpeername()[0] or (self.channel.path == 'report' and self.channel.datastore.downloaders.setdefault(self.headers['X-Request-Id'], self) != self):
        raise
    except:
      self.closed = True
      return

  def closed_callback(self):
    super().closed_callback()
    if self.channel.path == 'report' and self.channel.datastore.downloaders.pop(self.headers['X-Request-Id'], None) == self:
      progression = json.loads(self.channel.datastore.monitor_datastore.progress(self.headers['X-Request-Id']))
      progression['progress']['status'] = 'aborted'
      for sec in progression['progress'].get('sections', ()):
        sec['status'] = 'aborted'
      self.channel.datastore.monitor_datastore.progress = json.dumps(progression, separators=(',', ':'))
    self.server.close_event.set()


try:
  DownloadsWSServer = WebSocketIDAltServer(int(sys.argv[1]), DownloadsWSRequestHandler)
  DownloadsWSServer.start()
  sys.stderr.buffer.write(b'0')
  sys.stderr.buffer.flush()
except:
  sys.stderr.buffer.write(b'1')
  sys.stderr.buffer.flush()
  exit(1)
sys.stderr = open('con', 'w')
DownloadsWSServer.close_event = threading.Event()
DownloadsDSReport = DownloadsReportDS(DownloadsWSServer)
DownloadsWSServer.open('/report', DownloadsDSReport)
DownloadsWSServer.open('/monitor', DownloadsDSReport.monitor_datastore)
DownloadsWSServer.close_event.set()
while True:
  DownloadsWSServer.close_event.wait()
  DownloadsWSServer.close_event.clear()
  with DownloadsWSServer.lock:
    if any(not h.closed for c in DownloadsWSServer.channels.values() for h in c.handlers):
      continue
  if DownloadsWSServer.close_event.wait(10):
    continue
  with DownloadsWSServer.lock:
    if any(not h.closed for c in DownloadsWSServer.channels.values() for h in c.handlers):
      continue
    DownloadsWSServer.close('/report', timeout=1, block_on_close=True)
    DownloadsWSServer.close('/monitor', timeout=1, block_on_close=True)
  break
DownloadsWSServer.shutdown(timeout=1, block_on_close=True)


from SocketTB import *
import ssl
import base64
import os
import threading
import io
import time
import locale
import email.utils
import webbrowser
import msvcrt

with ISocketGenerator() as IGen:
  ISock0 = IGen(family=socket.AF_INET, type=socket.SOCK_STREAM)
  sock2 = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM)
  ISock2 = IGen(sock2)
  ISock0.bind(('127.0.0.1', 9000))
  ISock0.listen()
  ISock2.connect(('127.0.0.1', 9000))
  ISock1 = ISock0.accept()[0]
  print(ISock0, ISock1, ISock2)
  ISock2.sendall(b'test')
  print(ISock1.recv(10))
  try:
    ISock1.recv(10, timeout=2)
  except TimeoutError:
    print('timeout')
  def r(n):
    try:
      ISock1.recv(n)
    except OSError as err:
      if err.winerror == 10038:
        print('interrupted')
      else:
        raise
  t = threading.Thread(target=r, args=(10, ))
  t.start()
  time.sleep(2)
  ISock1.shutclose()
  def a():
    try:
      ISock1 = ISock0.accept()[0]
    except OSError as err:
      if err.winerror == 10038:
        print('interrupted')
      else:
        raise
  t = threading.Thread(target=a)
  t.start()
  time.sleep(2)

with IDSocketGenerator() as IDGen:
  IDSock0 = IDGen(family=socket.AF_INET, type=socket.SOCK_STREAM)
  sock2 = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM)
  IDSock2 = IDGen(sock2)
  IDSock0.bind(('127.0.0.1', 9000))
  IDSock0.listen()
  IDSock2.connect(('127.0.0.1', 9000))
  IDSock1 = IDSock0.accept()[0]
  print(IDSock0, IDSock1, IDSock2)
  IDSock2.sendall(b'test2')
  IDSock1.sendall(b'test1')
  print(IDSock1.recv(10))
  print(IDSock2.recv(10))
  def r1(n):
    try:
      IDSock1.recv(n)
    except OSError as err:
      if err.winerror == 10038:
        print('interrupted1')
      else:
        raise
  def r2(n):
    try:
      IDSock2.recv(n)
    except OSError as err:
      if err.winerror == 10038:
        print('interrupted2')
      else:
        raise
  t1 = threading.Thread(target=r1, args=(10, ))
  t2 = threading.Thread(target=r2, args=(10, ))
  t1.start()
  t2.start()
  time.sleep(2)

ctxs = NestedSSLContext(ssl.PROTOCOL_TLS_SERVER)
ctxs.load_autogenerated_cert_chain()
ctxc = NestedSSLContext(ssl.PROTOCOL_TLS_CLIENT)
ctxc.check_hostname = False
ctxc.verify_mode = ssl.CERT_NONE
with IDSocketGenerator() as IDGen:
  IDSock0 = IDGen(family=socket.AF_INET, type=socket.SOCK_STREAM)
  IDSock2 = IDGen(family=socket.AF_INET, type=socket.SOCK_STREAM)
  IDSock0S = ctxs.wrap_socket(IDSock0, server_side=True)
  IDSock0SS = ctxs.wrap_socket(IDSock0S, server_side=True)
  IDSock2S = ctxc.wrap_socket(IDSock2, server_side=False, server_hostname='127.0.0.1')
  IDSock2SS = ctxc.wrap_socket(IDSock2S, server_side=False, server_hostname='127.0.0.1')
  IDSock0SS.bind(('127.0.0.1', 9000))
  IDSock0SS.listen()
  def c():
    IDSock2SS.connect(('127.0.0.1', 9000))
  t = threading.Thread(target=c)
  t.start()
  IDSock1SS = IDSock0SS.accept()[0]
  print(IDSock0SS, IDSock1SS, IDSock2SS)
  IDSock2SS.sendall(b'test')
  print(IDSock1SS.recv(10))
  IDSock0SS.shutclose()
  IDSock1SS.shutclose()
  IDSock2SS.shutclose()
print(IDSock0SS, IDSock1SS, IDSock2SS)

UDPServer = MultiUDPIServer(1900, RequestHandler, True, ('239.255.255.250', 'ff02::c'))
UDPServer.start()
with ISocketGenerator() as IGen:
  ISock0 = IGen(family=socket.AF_INET, type=socket.SOCK_DGRAM)
  ISock0.sendto(b'test ipv4', ('239.255.255.250', 1900))
  ISock0.close()
  ISock0 = IGen(family=socket.AF_INET6, type=socket.SOCK_DGRAM)
  ISock0.sendto(b'test ipv6', ('ff02::c', 1900))
  ISock0.close()
  ISock0 = IGen(family=socket.AF_INET6, type=socket.SOCK_DGRAM)
  ISock0.bind(('::1',0))
  ISock0.sendto(b'test ipv6 lb', ('ff02::c', 1900))
time.sleep(2)
UDPServer.shutdown()

DS = WebSocketDataStore()
ctxs = NestedSSLContext(ssl.PROTOCOL_TLS_SERVER)
cid = base64.b32encode(os.urandom(10)).decode('utf-8')
with RSASelfSigned('TCPIServer' + cid, 1) as cert:
  cert.pipe_PEM('cert' + cid, 'key' + cid, 2)
  ctxs.load_cert_chain(r'\\.\pipe\cert%s.pem' % cid, r'\\.\pipe\key%s.pem' % cid)
WSSServer = WebSocketIDServer(('127.0.0.1', 9000), nssl_context=ctxs)
WSSServer.start()
WSSServer.open('/test')
WSSClient = WebSocketIDClient('wss://127.0.0.1:9000/test', DS)
WSSServer.broadcast('/test', 'welcome')
time.sleep(0.5)
print(DS.incoming)
WSSClient.close()
WSSServer.close('/test')
WSSServer.stop()

with IDSocketGenerator() as IDGen:
  HTTPRequest = HTTPRequestConstructor(IDGen)
  rep = HTTPRequest('https://www.python.org')
  print(rep)
  pcon = []
  rep = HTTPRequest('https://www.python.org', max_length=-1, pconnection=pcon)
  print(pcon)
  print(rep.body(20))
  print(rep.body(20))
  pcon[0].shutclose()
  rep = HTTPRequest('http://www.google.fr/search', max_length=-1, pconnection=pcon, process_cookies=True)
  print(pcon)
  print(rep.body(20))
  pcon[0].shutclose()

BA = HTTPBasicAuthenticator()
with HTTPIServer(9000, '', threaded=True, max_upload_size=10, dual_stack=True, basic_auth=BA), IDAltSocketGenerator() as IDGen:
  BA.set_realm('/', 'root')
  BA.set_credential('user', 'password', 'root')
  IDSock0 = IDGen(family=socket.AF_INET)
  IDSock0.connect(('127.0.0.1', 9000))
  IDSock0.send(b'OPTIONS * HTTP/1.1\r\nHost: 127.0.0.1:9000\r\n\r\n')
  print(IDSock0.recv(1024))
  IDSock0.close()
  IDSock0 = IDGen(family=socket.AF_INET6)
  IDSock0.connect(('[::1]', 9000))
  IDSock0.send(b'OPTIONS * HTTP/1.1\r\nHost: [::1]:9000\r\n\r\n')
  print(IDSock0.recv(1024))
  IDSock0.close()
  HTTPRequest = HTTPRequestConstructor(IDGen)
  pcon = []
  print(HTTPRequest('http://localhost:9000/*', method='OPTIONS', pconnection=pcon))
  print(HTTPRequest('http://localhost:9000/', pconnection=pcon))
  print(pcon)
  print(HTTPRequest('http://localhost:9000/', pconnection=pcon, basic_auth='user:password'))
  BA.set_realm('/', None)
  print(HTTPRequest('http://localhost:9000/nul', method='PUT', data=b'0' * 11, pconnection=pcon))
  print(pcon)
  u = HTTPIUpload('http://localhost:9000/nul', data=b'0' * 18)
  u.start()
  print(u.wait_finish(), u.progress['error'])
  u = HTTPIUpload('http://localhost:9000/nul', data=b'0' * 8)
  u.start()
  print(u.wait_finish())
  b = io.BytesIO()
  with HTTPIDownload('http://localhost:9000/test.py', b, max_workers=3, timeout=5, section_min=2, block_size=2) as d:
    while int(d.wait_progression().rstrip(' %')) <= 25:
      print(d.wait_sections(0))
    d.stop()
  print(d)
  with HTTPIDownload('http://localhost:9000/test.py', b, max_workers=2, timeout=5, section_min=4, block_size=4, resume=d) as d:
    while d.wait_finish(0) not in ('completed', 'aborted'):
      print(d.wait_sections())
    print(d)
  print(len(b.getvalue()))
  b.close()

locale.setlocale(locale.LC_TIME, '')
with NTPClient('time.google.com') as ntpc:
  print(ntpc.get_time(to_local=True))
  print(ntpc.get_offset())

with TOTPassword('AAAAAAAAAAAAAAAA') as totp:
  for i in range(10):
    p, r = totp.get(clipboard=True)
    print('', p, str(r).rjust(4), end ='\b'*12, flush=True)
    time.sleep(1)
  print('')

class TestRequestHandler(RequestHandler):
  def handle(self):
    if self.server.closed:
      return
    req = HTTPMessage(self.request)
    if req.method not in {'GET', 'HEAD'} or (body := bodies.get(req.path.lower())) is None:
      return
    resp = \
      'HTTP/1.1 200 OK\r\n' \
      'Date: %s\r\n' \
      'Server: SocketTB\r\n' \
      'Content-Length: %d\r\n' \
      'Content-Type: text/html; charset=utf-8\r\n' \
      'Connection: close\r\n' \
      '\r\n' % (email.utils.formatdate(time.time(), usegmt=True), len(body))
    self.request.sendall(resp.encode('ISO-8859-1'))
    if req.method != 'HEAD':
      self.request.sendall(body)
BA = WebRTCBasicAuthenticator()
code = BA.set_channel('test')
gist = 'https://gist.githubusercontent.com/PCigales/dc41dc6f499bfd49019b503857c0f57f/raw/'
for n in 'ABCer':
  BA.set_credential(n, 'WebRTC', 'test')
bodies = {}
for names in (('wrtc_mes.html',), ('wrtc_vid.html',), ('wrtc_capt_e.html', 'wrtc_capt_r.html')):
  with WebRTCSignalingServer(9000, basic_auth=BA) as SignalingServer, TCPIServer(9001, TestRequestHandler, threaded=True):
    SignalingServer.open('test')
    HTTPRequest = HTTPRequestConstructor(SignalingServer.idsocketgen)
    for name in names:
      print('http://127.0.0.1:9001/' + name)
      bodies['/' + name] = HTTPRequest(gist + name).body
      webbrowser.open('http://127.0.0.1:9001/' + name)
    print('Press "S" to stop...')
    while True:
      k = msvcrt.getch()
      if k == b'\xe0':
        k = msvcrt.getch()
        k = b''
      if k.upper() == b'S':
        break
    SignalingServer.close('test')
    bodies.clear()


# idownload v1.1 (https://github.com/PCigales/WinSocketTB)
# Copyright © 2025 PCigales
# This program is licensed under the GNU GPLv3 copyleft license (see https://www.gnu.org/licenses)

import sys
import subprocess
import os.path


if len(sys.argv) <= 1:
  import winreg
  for k, m in zip((r'SOFTWARE\Mozilla', r'SOFTWARE\Microsoft\Edge'), ('idownload_F.json', 'idownload_E.json')):
    try:
      winreg.QueryValue(winreg.HKEY_CURRENT_USER, k)
      sk = k + r'\NativeMessagingHosts\idownload'
      winreg.SetValue(winreg.HKEY_CURRENT_USER, sk, winreg.REG_SZ, os.path.join(os.path.dirname(os.path.abspath(globals().get('__file__', ' '))), m))
      print('HKEY_CURRENT_USER\\' + sk, '=', winreg.QueryValue(winreg.HKEY_CURRENT_USER, sk))
    except:
      pass
  exit(0)

process = subprocess.Popen(('py', os.path.join(os.path.dirname(os.path.abspath(globals().get('__file__', ' '))), 'downloader.py')), creationflags=(subprocess.CREATE_BREAKAWAY_FROM_JOB | subprocess.CREATE_NEW_CONSOLE), stdin=sys.stdin, stdout=sys.stdout, stderr=subprocess.PIPE)
process.stderr.read(1)
exit(0)